from blockapi.services import BlockchainAPI


class ChainSoAPI(BlockchainAPI):
    """
    Multi coins: bitcoin, litecoin, dogecoin, zcash, dash
    Does't support xpub/ypub/zpub :(
    API docs: https://chain.so/api
    Explorer: 
    """

    symbol = None
    base_url = 'https://chain.so/api/v2'
    rate_limit = 0.2  # 5 per second
    coef = None
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/get_address_balance/{symbol}/{address}',
        'get_txs': '/address/{symbol}/{address}'
    }

    def __init__(self, address, api_key=None):
        super().__init__(address, api_key)

    def get_balance(self):
        response = self._request('get_balance')
        if not response:
            return None

        retval = float(response['confirmed_balance']) * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]

    # don't set default args, we can get all transactions at once
    # @set_default_args_values
    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        response = self._request('get_txs')
        if not response:
            return []

        txs = response['txs']
        # filter records manually because from api we get all txs
        if offset or limit:
            txs = txs[offset:limit]

        return [self.parse_tx(t) for t in txs]

    @staticmethod
    def parse_tx(tx):
        return tx

    def _request(self, method):
        response = self.request(
            method,
            symbol=self.symbol,
            address=self.address,
            with_cloudflare=True
        )
        if response['status'] == 'fail':
            return None
        return response['data']


class ChainSoBitcoinAPI(ChainSoAPI):
    active = False
    symbol = 'BTC'
    coef = 1e-8


class ChainSoLitecoinAPI(ChainSoAPI):
    active = False
    symbol = 'LTC'
    coef = 1e-8


class ChainSoDogecoinAPI(ChainSoAPI):
    symbol = 'DOGE'
    coef = 1


class ChainSoZcashAPI(ChainSoAPI):
    symbol = 'ZEC'
    coef = 1


class ChainSoDashAPI(ChainSoAPI):
    symbol = 'DASH'
    coef = 1



from blockapi.services import BlockchainAPI


class GreymassAPI(BlockchainAPI):
    """
    coins: eos
    API docs: https://github.com/greymass/eosio-api-ext/wiki
    /API-Request-Response-Examples
    Explorer: 
    """

    active = True

    symbol = 'EOS'
    base_url = 'https://eos.greymass.com/v1'
    rate_limit = 0
    coef = 1
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/chain/get_currency_balances',
    }

    def get_balance(self):
        body = '{"account": "' + self.address + '"}'
        response = self.request('get_balance',
                                body=body)
        if not response:
            return None

        # {‘symbol’: _, ‘address’: _, ‘price’: _, ‘name’: ?} ]

        try:
            return [{"symbol": item["symbol"],
                     "address": item["code"], 
                     "amount": float(item["amount"]),
                     "name": None} for item in response]
        except KeyError:
            return None



from pytest import mark

from blockapi.api.btc import BtcAPI
from blockapi.test_init import test_addresses


class TestBtcAPI:
    ADDRESS = test_addresses['BCH'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = BtcAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'BCH', 'amount': 505295.26508585}]



from blockapi.services import BlockchainAPI


class StellarAPI(BlockchainAPI):
    """
    coins: stellar
    API docs: https://www.stellar.org/developers/horizon/reference/index.html
    Explorer: https://stellarchain.io/
    """

    active = True

    symbol = 'XLM'
    base_url = 'https://horizon.stellar.org'
    rate_limit = 0
    coef = 1
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/accounts/{address}',
    }

    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        balances = [{
            'symbol': bal['asset_code'],
            'amount': float(bal['balance'])
        } for bal in response['balances'] if 'asset_code' in bal]
        balance_stellar = [{
            'symbol': 'XLM',
            'amount': float(bal['balance'])
        } for bal in response['balances'] if bal['asset_type'] == 'native']

        return balances + balance_stellar



from pytest import mark

from blockapi.api.tronscan import TronscanAPI
from blockapi.test_init import test_addresses


class TestTronscanAPI:
    ADDRESS = test_addresses['TRX'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = TronscanAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert next((r["amount"] for r in result if r["symbol"] == "TRX")) ==\
            0.588285
        assert len(result) == 45



from pytest import mark

from blockapi.api.insight import RavencoinAPI
from blockapi.test_init import test_addresses


class TestInsightAPI:
    RAVEN_ADDRESS = test_addresses['RVN'][1]

    @mark.vcr()
    def test_get_balance_raven_coin(self):
        api = RavencoinAPI(address=self.RAVEN_ADDRESS)
        result = api.get_balance()

        assert result == 957.0026321



from pytest import mark, raises

from blockapi.api.dcrdata import DcrdataAPI
from blockapi.services import InternalServerError, AddressNotExist
from blockapi.test_init import test_addresses


class TestDcrdataAPI:

    ADDRESS = test_addresses["DCR"][0]
    STAKING_ADDRESS = test_addresses["DCR"][1]

    def test_init(self):
        api = DcrdataAPI(address=self.ADDRESS)
        assert api

    def test_process_error_response(self):
        response = TestResponse()

        response.text = "Error"
        response.status_code = 500

        with raises(InternalServerError):
            DcrdataAPI(address=self.ADDRESS).process_error_response(response)

        response.status_code = 422

        with raises(AddressNotExist):
            DcrdataAPI(address=self.ADDRESS).process_error_response(response)

    @mark.vcr()
    def test_get_balance(self):
        api = DcrdataAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'DCR', 'amount': 135.0176}]

    @mark.vcr()
    def test_get_txs(self):
        api = DcrdataAPI(address=self.ADDRESS)
        result = api.get_txs(limit=2)

        assert len(result) == 2
        assert all(k in ["kind", "result"] for k in result[0].keys())

    @mark.vcr()
    def test_get_txs_staking(self):
        api = DcrdataAPI(address=self.STAKING_ADDRESS)
        result = api.get_txs()

        assert isinstance(result, list)
        assert len(result) == 21
        assert result[0]["kind"] == "ticket"
        assert result[0]["result"]["hash"] == "51f27646cd8b98873816d820281bb"\
            "bf850de79564440e3b18eceef46ef56cfae"
        assert result[0]["result"]["purchased_on"].year == 2020
        assert result[0]["result"]["purchased_on"].month == 6
        assert result[0]["result"]["purchased_on"].day == 5
        assert result[0]["result"]["investment"] == 146.59688866
        assert result[0]["result"]["ticket_cost"] == 146.59683446
        assert result[0]["result"]["transaction_fee"] == 5.419999999389802e-05
        assert result[0]["result"]["pool_fee"] == 0.0087343
        assert result[0]["result"]["status"] == "live"

    @mark.vcr()
    def test_get_tx(self):
        hash_ = "193b59d8926588181aad5a5bed672e1fccf443f4d05dc1" \
                "6a0cdaacf3b4b4ed7c"
        api = DcrdataAPI(address=self.ADDRESS)
        result = api.get_tx(tx_hash=hash_)

        assert len(result) == 1


class TestResponse:

    text = None
    status_code = 200



from blockapi.services import AddressNotExist, BlockchainAPI


class BlockcypherAPI(BlockchainAPI):
    """
    Multi coins: bitcoin, litecoin, dogecoin
    API docs: https://github.com/Blockchair/Blockchair.Support/blob
    /master/API_DOCUMENTATION_EN.md
    Explorer: https://live.blockcypher.com
    """

    symbol = None
    base_url = 'https://api.blockcypher.com/v1/'
    rate_limit = 0
    coef = None
    start_offset = 0
    max_items_per_page = None
    page_offset_step = max_items_per_page

    supported_requests = {
        # for limit and offset the second parameter 0 is for utxo
        'get_balance': '/{coin}/main/addrs/{address}',
    }

    def process_error_response(self, response):
        err = response.json().get('error')
        if 'is invalid' in err or err == 'Error: wallet not found':
            raise AddressNotExist()
        # else
        super().process_error_response(response)

    def get_balance(self):
        response = self.request(
            'get_balance',
            coin=self.symbol.lower(),
            address=self.address
        )
        if not response:
            return None

        retval = response['balance'] * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]


class BlockcypherLitecoinAPI(BlockcypherAPI):
    symbol = 'LTC'
    coef = 1e-8



from datetime import datetime

import pytz

from blockapi.services import BlockchainAPI


class BlockonomicsAPI(BlockchainAPI):
    """
    Bitcoin
    API docs: https://www.blockonomics.co/views/api.html
    Explorer: https://www.blockonomics.co
    """

    active = False

    symbol = 'BTC'
    base_url = 'https://www.blockonomics.co/api'
    rate_limit = 30
    coef = 1e-8
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None
    xpub_support = True

    supported_requests = {
        'get_balance': '/balance',
        'get_txs': '/searchhistory',
        'get_tx': '/tx_detail?txid={txid}'
    }

    def __init__(self, address, api_key):
        super().__init__(address=address, api_key=api_key)

    def get_balance(self):
        body = '{"addr": "' + self.address + '"}'
        response = self.request('get_balance', body=body)
        if not response.get('response'):
            return None

        balance = sum(r['confirmed'] * self.coef for r in response['response'])
        return [{'symbol': self.symbol, 'amount': balance}]

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        body = '{"addr": "' + self.address + '"}'
        txs = self.request('get_txs',
                           body=body)

        return [self.parse_tx(t) for t in txs['history']]

    def parse_tx(self, tx):
        tx_data = self.request('get_tx', 
                               txid=tx['txid'])

        amount = tx['value']
        from_address = [adr['address'] for adr in tx_data['vin']]
        to_address = [adr['address'] for adr in tx_data['vout']]

        if amount < 0:
            direction = 'outgoing'
        else:
            direction = 'incoming'

        return {
            'date': datetime.fromtimestamp(tx['time'], pytz.utc),
            'from_address': from_address,
            'to_address': to_address,
            'amount': amount * self.coef,
            'fee': tx_data['fee'] * self.coef,
            'gas': {},
            'hash': tx['txid'],
            'confirmed': None,
            'is_error': False,
            'type': 'normal',
            'kind': 'transaction',
            'direction': direction,
            'raw': tx 
        }



from pytest import mark

from blockapi.api.digonchain import DigonchainAPI
from blockapi.test_init import test_addresses


class TestDigonchainAPI:
    ADDRESS = test_addresses['VET'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = DigonchainAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'VET', 'amount': 553093.54494999}]



from pytest import mark

from blockapi.api.blockchainos import BlockchainosAPI
from blockapi.test_init import test_addresses


class TestBlockchainosAPI:
    ADDRESS = test_addresses['BOS'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = BlockchainosAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'BOS', 'amount': 454146425.404}]

    @mark.vcr()
    def test_get_txs(self):
        api = BlockchainosAPI(address=self.ADDRESS)
        result = api.get_txs(limit=10)

        assert isinstance(result, list)
        assert len(result) == 10



from blockapi.api.alethio import *
from blockapi.api.binance import *
from blockapi.api.blockchaininfo import *
from blockapi.api.blockchainos import *
from blockapi.api.blockchair import *
from blockapi.api.blockcypher import *
from blockapi.api.blockonomics import *
from blockapi.api.blockscout import *
from blockapi.api.btc import *
from blockapi.api.cardanoexplorer import *
from blockapi.api.chainso import *
from blockapi.api.cosmos import *
from blockapi.api.cryptoid import *
from blockapi.api.dcrdata import *
from blockapi.api.digonchain import *
from blockapi.api.eospark import *
from blockapi.api.etherscan import *
from blockapi.api.ethplorer import *
from blockapi.api.greymass import *
from blockapi.api.insight import *
from blockapi.api.neoscan import *
from blockapi.api.ontology import *
from blockapi.api.stellar import *
from blockapi.api.trezor import *
from blockapi.api.tronscan import *
from blockapi.api.tzscan import *
from blockapi.api.zchain import *
from blockapi.api.zensystem import *



import inspect
import random

import coinaddrng

import blockapi.api
import blockapi.utils
from .services import APIError
from .test_init import test_addresses

# currencies' ids (from coingecko.com) and symbols
COINS = {
    'binance-coin': 'BNB',
    'bitcoin': 'BTC',
    'bitcoin-cash': 'BCH',
    'bitcoin-sv': 'BSV',
    'boscoin': 'BOS',
    'cardano': 'ADA',
    'cosmos': 'ATOM',
    'decred': 'DCR',
    'dashcoin': 'DASH',
    'dogecoin': 'DOGE',
    'eos': 'EOS',
    'ethereum': 'ETH',
    'ethereum-classic': 'ETC',
    'groestlcoin': 'GRS',
    'horizen': 'ZEN',
    'litecoin': 'LTC',
    'neocoin': 'NEO',
    'ontology': 'ONT',
    'ravencoin': 'RVN',
    'stellar': 'XLM',
    'tezos': 'XTZ',
    'tron': 'TRX',
    'vechain': 'VET',
    'zcash': 'ZEC'
}


def get_balance_from_random_api(symbol, address):
    """Get balance for currency from random API
    (APIs with API keys are not supported).
    """
    return _call_method_from_random_api(symbol, address, 'get_balance')


def get_shuffled_suitable_api_classes_for_coin(symbol, address):
    api_classes = get_shuffled_api_classes_for_coin(symbol)
    filtered_api_classes = filter_suitable_api_classes(api_classes, symbol,
                                                       address)
    return filtered_api_classes


def _call_method_from_random_api(symbol, address, method):
    filtered_api_classes = get_shuffled_suitable_api_classes_for_coin(symbol,
                                                                      address)
    for cl in filtered_api_classes:
        try:
            inst = cl(address)
            return getattr(inst, method)()
        except (APIError, Exception):
            continue
    return None


def get_shuffled_api_classes_for_coin(symbol):
    api_classes = get_api_classes_for_coin(symbol)
    random.shuffle(api_classes)
    return api_classes


def get_api_classes_for_coin(symbol):
    return [i for i in get_active_api_classes() if
            i.symbol and i.symbol == symbol]


def filter_suitable_api_classes(api_classes, symbol, address):
    address_info = get_address_info(symbol, address)

    filtered = api_classes
    if address_info.network == 'test':
        filtered = [c for c in filtered if c.testnet_url]
    if address_info.is_extended:
        filtered = [c for c in filtered if c.xpub_support]
    return filtered


def get_random_api_class_for_coin(symbol, exclude=None):
    api_classes = get_api_classes_for_coin(symbol)
    exclude = [] if not exclude else exclude
    api_classes = [cl for cl in api_classes if not (cl in exclude)]
    return random.choice(api_classes) if api_classes else None


def get_all_supported_coins():
    return list(set(c.symbol for c in get_active_api_classes()
                    if c.symbol))


def get_active_api_classes():
    inheritors = _get_all_inheritors()
    return [i for i in inheritors if i.active]


def _inheritors(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


def _get_subclasses(class_name):
    return [getattr(class_name, x) for x in dir(class_name)
            if not x.startswith('__')]


def _get_all_inheritors():
    all_inheritors = []

    tridy = _get_subclasses(blockapi.api)

    for trida in tridy:
        tridy_sub = _get_subclasses(trida)
        for trida_sub in tridy_sub:
            if (inspect.isclass(trida_sub) and
                    blockapi.services.BlockchainAPI in trida_sub.__bases__):
                all_inheritors.append(trida_sub)
                grandchildren = _inheritors(trida_sub)
                if len(grandchildren) > 0:
                    all_inheritors += _inheritors(trida_sub)

    return all_inheritors


def get_working_apis_for_coin(symbol, debug=False):
    coin_classes = get_api_classes_for_coin(symbol)

    if symbol in test_addresses:
        if len(test_addresses[symbol]) == 0:
            return ()
    else:
        return ()

    coin_address = test_addresses[symbol][0]
    working_apis = []

    api_ok = True
    for api_class in coin_classes:
        api_inst = api_class(coin_address)
        exception_class = None
        exception_msg = None
        try:
            api_inst.get_balance()

        except Exception as e:
            exception_class = e.__class__.__name__
            exception_msg = str(e)
            api_ok = False

        if debug:
            working_apis.append((api_class, exception_class, exception_msg))
        else:
            if api_ok:
                working_apis.append(api_class)

    return tuple(working_apis)


def get_working_apis(debug=False):
    all_coins = get_all_supported_coins()
    ok_apis = {}

    for a_coin in all_coins:
        ok_apis[a_coin] = get_working_apis_for_coin(a_coin, debug=debug)

    return ok_apis


def check_address_valid(symbol, address):
    return get_address_info(symbol, address).valid


def get_address_info(symbol: str, address: str):
    try:
        return coinaddrng.validate(symbol.lower(), address)
    except TypeError:
        # if validator for symbol doesn't exist return default object;
        # 'valid' attribute is set to True, because there may not exist
        #  validator for every supported coin
        return coinaddrng.validation.ValidationResult(
            name='',
            ticker=symbol,
            address=address.encode(),
            valid=True,
            network='',
            address_type='',
            is_extended=False
        )



import inspect
from abc import ABC, abstractmethod
from datetime import datetime
from time import sleep

import cfscrape
import requests

import blockapi

cfscrape.DEFAULT_CIPHERS += ':!SHA'


class Service(ABC):
    """General class for handling blockchain API services."""

    active = True

    base_url = None
    rate_limit = 0

    # {request_method: request_url}
    supported_requests = {}

    def __init__(self, api_key=None):
        self.api_key = api_key
        self.last_response = None
        self.last_response_time = None

    def build_request_url(self, request_method, **params):
        path_url = self.supported_requests.get(request_method)
        if path_url:
            return self.base_url + path_url.format(**params)
        return None

    def request(self, request_method, with_rate_limit=False,
                with_cloudflare=False, body=None, headers=None, **params):
        request_url = self.build_request_url(request_method, **params)

        if not request_url:
            return None

        if not body:
            body = {}

        if not headers:
            headers = {}

        if with_cloudflare:
            reqobj = cfscrape.create_scraper()
        else:
            reqobj = requests

        if with_rate_limit and self.rate_limit:
            self.wait_for_next_request()

        # if body is passed, use post
        if body:
            response = reqobj.post(request_url, data=body, headers=headers)
        else:
            response = reqobj.get(request_url)

        self.last_response = response
        self.last_response_time = datetime.now()

        if response.status_code != 200:
            self.process_error_response(response)

        return response.json()

    def wait_for_next_request(self):
        if not self.last_response_time:
            return

        diff = (datetime.now() - self.last_response_time).total_seconds()
        wait = self.rate_limit - diff

        if wait > 0:
            sleep(wait)

        # use date from last response
        # doesn't work very good, time on server can differ from local time
        # if self.last_response.headers.get('Date'):
        # last_resp_time = date_parse(self.last_response.headers.get('Date'))
        # last_resp_time.replace(tzinfo=UTC)
        # wait_until = last_resp_time + timedelta(seconds=self.rate_limit)
        #
        # now = datetime.utcnow().replace(tzinfo=UTC)

    def process_error_response(self, response):
        if response.status_code == 500:
            raise InternalServerError('Error 500: Internal Server Error.')
        if response.status_code == 502:
            raise BadGateway('Error 502: Bad Gateway.')
        if response.status_code == 504:
            raise GatewayTimeOut('Error 504: Gateway timeout.')

        raise APIError('Following error occured: {}, status code: {}.'.format(
            response.text, response.status_code))


# Decorator that set default args
def set_default_args_values(f):
    def wrapper(*args, **kwargs):
        args, kwargs = _set_default_arg(
            f, 'offset', args[0].start_offset, *args, **kwargs)
        args, kwargs = _set_default_arg(
            f, 'limit', args[0].max_items_per_page, *args, **kwargs)

        return f(*args, **kwargs)

    def _set_default_arg(f_, arg_name, default_value, *args, **kwargs):
        sig = inspect.signature(f_)
        arg_idx, _ = next((i, par) for i, (name, par)
                          in enumerate(sig.parameters.items())
                          if name == arg_name)

        # check if its argument is in *args
        if len(args) > arg_idx:
            if args[arg_idx] is None:
                list_args = list(args)
                list_args[arg_idx] = default_value
                args = tuple(list_args)

        elif kwargs.get(arg_name) is None:
            kwargs[arg_name] = default_value

        return args, kwargs

    return wrapper


def on_failure_return_none():
    def decorate(f):
        def applicator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except (APIError,
                    InternalServerError,
                    BadGateway,
                    GatewayTimeOut):
                return None

        return applicator
    return decorate


# Exceptions
class APIError(Exception):
    pass


class AddressNotExist(APIError):
    pass


class APIKeyMissing(APIError):
    pass


# 500
class InternalServerError(APIError):
    pass


# 502
class BadGateway(APIError):
    pass


# 504
class GatewayTimeOut(APIError):
    pass


class BlockchainInterface(ABC):
    coef = 1
    start_offset = 0
    max_items_per_page = None
    page_offset_step = 1
    confirmed_num = 0
    testnet_url = None
    xpub_support = False

    def __init__(self, address):
        self.address = address

    @abstractmethod
    def get_balance(self):
        raise NotImplementedError()

    @set_default_args_values
    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        """Loads txs for single address."""
        return []

    @staticmethod
    def filter_unconfirmed_txs(txs):
        return [t for t in txs if t.get('confirmed') in [True, None]]


class BlockchainAPI(Service, BlockchainInterface, ABC):
    symbol = None

    def __init__(self, address, api_key=None):
        Service.__init__(self, api_key)
        BlockchainInterface.__init__(self, address)

        self.address_info = blockapi.get_address_info(self.symbol.lower(),
                                                      address)
        self.update_network()

    def update_network(self):
        if self.address_info.network == 'test':
            if self.testnet_url:
                self.base_url = self.testnet_url
            else:
                raise ValueError("API doesn't support testnet.")



from pytest import mark

from blockapi.api.binance import BinanceAPI

from blockapi.test_init import test_addresses


class TestBinanceAPI:

    ADDRESS = test_addresses["BNB"][0]

    def test_init(self):
        api = BinanceAPI(address=self.ADDRESS)
        assert api

    @mark.vcr()
    def test_get_balance(self):
        api = BinanceAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert isinstance(result, list)
        assert len(result) == 68

    @mark.vcr()
    def test_get_txs(self):
        api = BinanceAPI(address=self.ADDRESS)
        result = api.get_txs()

        assert isinstance(result, list)
        assert len(result) == 100



from datetime import datetime

import dateutil.parser
import pytz

from blockapi.services import (
    BlockchainAPI,
    set_default_args_values,
    AddressNotExist,
)


class BlockchairAPI(BlockchainAPI):
    """
    Multi coins: bitcoin, bitcoin-cash, bitcoin-sv, litecoin, dogecoin,
                 dash, ethereum, groestlcoin
    API docs: https://github.com/Blockchair/Blockchair.Support/blob
    /master/API_DOCUMENTATION_EN.md
    Explorer: https://blockchair.com
    """

    active = True

    base_url = 'https://api.blockchair.com'
    symbol = None
    name = None
    rate_limit = 0
    coef = None
    start_offset = 0
    max_items_per_page = 10  # 10000 per tx hashes; 10 per tx details
    page_offset_step = max_items_per_page
    xpub_support = True

    supported_requests = {
        # for limit and offset the second parameter 0 is for utxo
        'get_dashboard': '/{name}/dashboards/{address_type}'
                         '/{address}?limit={limit},0&offset={offset},0',
        'get_txs': '/{name}/dashboards/transactions/{hash_or_hashes}',
    }

    def __init__(self, address, api_key=None):
        super().__init__(address, api_key)
        self._set_address_type()

    def _set_address_type(self):
        is_xpub = (any(self.address.startswith(p)
                       for p in ['xpub', 'ypub', 'zpub']))
        self.address_type = 'xpub' if is_xpub else 'address'

    def get_balance(self):
        dashboard = self._get_dashboard()
        if not dashboard:
            return None

        retval = int(dashboard['address']['balance']) * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]

    def get_create_date(self):
        dashboard = self._get_dashboard()
        if not dashboard:
            return 0

        date_str = dashboard[self.address_type]['first_seen_receiving']
        date = datetime.strptime(date_str, '%Y-%m-%d %H:%M:%S')
        return date.replace(tzinfo=pytz.UTC)

    @set_default_args_values
    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        dashboard = self._get_dashboard(offset, limit)
        if not dashboard:
            return []

        tx_hashes = dashboard['transactions']
        tx_response = self.request(
            'get_txs',
            symbol=self.symbol,
            hash_or_hashes=','.join(tx_hashes)
        )
        if not tx_response.get('data'):
            return []

        txs = list(tx_response['data'].values())
        return [self.parse_tx(t) for t in txs]

    def parse_tx(self, tx):
        my_input = next((i for i in tx['inputs']
                         if self.address == i['recipient']), None)
        my_output = next((o for o in tx['outputs']
                          if self.address == o['recipient']), None)
        tx_data = tx['transaction']

        if my_input:
            amount = my_input['value'] * self.coef
            direction = 'outgoing'
            from_address = self.address
            to_address = (tx['outputs'][0]['recipient']
                          if tx_data['output_count'] == 1 else 'multiple')
        else:
            amount = my_output['value'] * self.coef
            direction = 'incoming'
            to_address = self.address
            from_address = (tx['inputs'][0]['recipient']
                            if tx_data['input_count'] == 1 else 'multiple')

        return {
            'date': dateutil.parser.parse(tx_data['time']),
            'from_address': from_address,
            'to_address': to_address,
            'amount': amount,
            'fee': tx_data['fee'] * self.coef,
            'gas': {},
            'hash': tx_data['hash'],
            'confirmed': None,
            'is_error': False,
            'type': 'normal',
            'kind': 'transaction',
            'direction': direction,
            'raw': tx
        }

    def _get_dashboard(self, offset=0, limit=0):
        response = self.request(
            'get_dashboard',
            symbol=self.symbol,
            name=self.name,
            address_type=self.address_type,
            address=self.address,
            offset=offset,
            limit=limit
        )

        data = response.get('data')
        if not data:
            raise AddressNotExist()

        dashboard = list(data.values())[0]

        if self.address_type == 'address' and not dashboard['address']['type']:
            raise AddressNotExist()

        return dashboard


class BlockchairBitcoinAPI(BlockchairAPI):
    symbol = 'BTC'
    name = 'bitcoin'
    coef = 1e-8
    active = False


class BlockchairBitcoinCashAPI(BlockchairAPI):
    symbol = 'BCH'
    name = 'bitcoin-cash'
    coef = 1e-8
    active = False


class BlockchairBitcoinSvAPI(BlockchairAPI):
    symbol = 'BSV'
    name = 'bitcoin-sv'
    coef = 1e-8


class BlockchairLitecoinAPI(BlockchairAPI):
    symbol = 'LTC'
    name = 'litecoin'
    coef = 1e-8
    active = False


class BlockchairDogecoinAPI(BlockchairAPI):
    symbol = 'DOGE'
    name = 'dogecoin'
    coef = 1e-8
    active = False


class BlockchairDashAPI(BlockchairAPI):
    symbol = 'DASH'
    name = 'dash'
    coef = 1e-8
    active = False


class BlockchairEthereumAPI(BlockchairAPI):
    symbol = 'ETH'
    name = 'ethereum'
    coef = 1e-18
    active = False


class BlockchairGroestlcoinAPI(BlockchairAPI):
    symbol = 'GRS'
    name = 'groestlcoin'
    coef = 1e-8



from blockapi.services import BlockchainAPI


class DigonchainAPI(BlockchainAPI):
    """
    coins: vechain
    API docs: https://vethor-pubnode.digonchain.com/doc/swagger-ui/
    Explorer: https://explorer.vtho.net/
    """

    active = True

    symbol = 'VET'
    # the docs say this node may not be reliable enough
    # but seems there's no other
    # TODO: find a better one in the future
    base_url = 'https://vethor-pubnode.digonchain.com'
    rate_limit = 0
    coef = 1e-18
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/accounts/{address}',
    }

    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None
        else:
            try:
                balance = response['balance']
                balance = int(balance, 16)
                retval = balance * self.coef
            except (KeyError, ValueError):
                return None

        return [{'symbol': self.symbol, 'amount': retval}]



import re
from copy import deepcopy

import dateutil.parser

from blockapi.services import (
    BlockchainAPI,
    APIError,
    set_default_args_values,
    AddressNotExist
)


class CosmosAPI(BlockchainAPI):
    """
    Cosmos
    API docs: https://cosmos.network/rpc/
    Explorer: https://www.mintscan.io
    """

    symbol = 'ATOM'
    # base_url is 'https://stargate.cosmos.network'
    base_url = 'https://cosmos.node.cracklord.com'
    rate_limit = 0.6
    coef = 1e-6
    start_offset = 1
    max_items_per_page = 30
    page_offset_step = 1

    supported_requests = {
        'get_info': '/auth/accounts/{address}',
        'get_balance': '/bank/balances/{address}',
        'get_txs': '/txs?action={action}&{role}={address}&page={page}'
                   '&limit={limit}',
        'get_delegations': '/staking/delegators/{address}/delegations',
        'get_unbonding_delegations': '/staking/delegators/{address}'
                                     '/unbonding_delegations',
        'get_redelegations': '/staking/redelegations?delegator={address}',
        'get_delegation_reward': '/distribution/delegators/{address}'
                                 '/rewards/{validator_address}',
        'get_proposals': '/gov/proposals',
        'query_votes': '/gov/proposals/{proposal_id}/votes',
    }

    def process_error_response(self, response):
        if 'decoding bech32 failed' in response.text:
            raise AddressNotExist()
        # else
        super().process_error_response(response)

    def get_info(self):
        return self.request('get_info', address=self.address)

    def get_balance(self):
        balances = self.request('get_balance', address=self.address)
        if not balances:
            return None

        balances_result = []
        for b in balances['result']:
            symbol = self.symbol if b['denom'] == 'uatom' else b['denom']
            balances_result.append({
                'symbol': symbol,
                'amount': int(b['amount']) * self.coef
            })
        return balances_result

    def get_delegations(self):
        return self.request('get_delegations', address=self.address)

    def get_unbonding_delegations(self):
        return self.request('get_unbonding_delegations', address=self.address)

    def get_redelegations(self):
        return self.request('get_redelegations', address=self.address)

    def get_delegation_reward(self, validator_address):
        return self.request('get_delegation_reward',
                            address=self.address,
                            validator_address=validator_address)

    def get_proposals(self):
        return self.request('get_proposals')

    def get_votes_deprecated(self):
        votes = []
        for prop in self.get_proposals():
            try:
                vote = self.request('query_votes',
                                    proposal_id=prop['proposal_id'])
                if vote is None:
                    continue
                for v in vote:
                    if v['voter'] == self.address:
                        votes.append(v)

            except APIError:
                pass

        return votes

    def get_incoming_txs(self, offset=None, limit=None, unconfirmed=False):
        txs = self._get_txs('send', 'recipient', offset, limit, unconfirmed)
        return self.parse_txs(txs)

    def get_outgoing_txs(self, offset=None, limit=None, unconfirmed=False):
        txs = self._get_txs('send', 'sender', offset, limit, unconfirmed)
        return self.parse_txs(txs)

    def get_multi_incoming_txs(self, offset=None, limit=None,
                               unconfirmed=False):
        txs = self._get_txs('multisend', 'recipient', offset, limit,
                            unconfirmed)
        return self.parse_txs(txs)

    def get_multi_outgoing_txs(self, offset=None, limit=None,
                               unconfirmed=False):
        txs = self._get_txs('multisend', 'sender', offset, limit, unconfirmed)
        return self.parse_txs(txs)

    def parse_txs(self, txs):
        parsed_txs = []
        for tx in txs['txs']:
            # there can be multiple txs in single tx
            for parsed in self.parse_tx(tx):
                parsed_txs.append(parsed)
        return parsed_txs

    def parse_tx(self, tx):
        fee_amount = int(tx['tx']['value']['fee']['amount'][0]['amount'])
        base_tx = {
            'date': dateutil.parser.parse(tx['timestamp']),
            'fee': fee_amount * self.coef if fee_amount else None,
            'gas': {
                'gas_used': int(tx['gas_used']),
                'gas_limit': int(tx['gas_wanted'])
            },
            'hash': tx['txhash'],
            'confirmed': None,
            'type': 'normal',
            'kind': 'transaction',
            'description': tx['tx']['value']['memo'],
            'raw': tx
        }

        # for every message (sub tx?) create new tx
        msgs = []
        for msg in self._parse_messages(tx):
            msg['direction'] = (
                'incoming'
                if msg['to_address'].lower() == self.address.lower()
                else 'outgoing'
            )
            msg.update(base_tx)
            msgs.append(msg)
        return msgs

    #
    def get_rewards_withdrawals(self, offset=None, limit=None):
        items = self._get_txs('withdraw_delegator_reward', 'delegator', offset,
                              limit)
        return self.parse_other_txs(items)

    #
    def get_redelegates(self, offset=None, limit=None):
        items = self._get_txs('begin_redelegate', 'delegator', offset, limit)
        return self.parse_other_txs(items)

    #
    def get_delegates(self, offset=None, limit=None):
        items = self._get_txs('delegate', 'delegator', offset, limit)
        return self.parse_other_txs(items)

    #
    def get_undelegates(self, offset=None, limit=None):
        items = self._get_txs('begin_unbonding', 'delegator', offset, limit)
        return self.parse_other_txs(items)

    def get_votes(self, offset=None, limit=None):
        items = self._get_txs('vote', 'voter', offset, limit)
        return self.parse_other_txs(items)

    # only for validators
    def get_validator_edits(self, offset=None, limit=None):
        items = self._get_txs('edit_validator', 'destination-validator',
                              offset, limit)
        return self.parse_other_txs(items)

    # create validator
    # https://stargate.cosmos.network/txs/938AFA947B6EBE4CA614FFD7C8F98F2452
    # 8A395061FAC68EF1EFB92422F6158F

    def parse_other_txs(self, txs):
        parsed_txs = []
        for tx in txs:
            # there can be multiple txs in single tx
            for parsed in self._parse_other_tx(tx):
                parsed_txs.append(parsed)
        return parsed_txs

    @set_default_args_values
    def _get_txs(self, action, role, offset=None, limit=None,
                 unconfirmed=None):
        return self.request(
            'get_txs',
            action=action,
            role=role,
            address=self.address,
            page=offset,
            limit=limit,
            unconfirmed=unconfirmed
        )

    def _parse_other_tx(self, tx):
        fee_data = tx['tx']['value']['fee']['amount']
        fee_amount = int(fee_data[0]['amount']) * self.coef if fee_data \
            else None

        base_tx = {
            'date': dateutil.parser.parse(tx['timestamp']),
            'fee': fee_amount,
            'gas': {
                'gas_used': int(tx['gas_used']),
                'gas_limit': int(tx['gas_wanted'])
            },
            'hash': tx['txhash'],
            'description': tx['tx']['value']['memo'],
            'raw': tx
        }

        # for every message (sub tx?) create new tx
        msgs = []
        for msg in self._parse_messages(tx):
            _tx = deepcopy(base_tx)
            _tx['kind'] = msg.pop('kind')
            _tx['operation'] = msg
            msgs.append(_tx)
        return msgs

    def _parse_messages(self, tx):
        logs = deepcopy(tx['logs'])

        # only reward has values in separated dict
        r = re.compile("([0-9]+)([a-zA-Z]+)")
        # [($number, $coin), ...]
        rewards = [r.match(t['value']).groups()
                   for t in tx['tags'] if t['key'] == 'rewards']

        parsed_msgs = []
        for msg in tx['tx']['value']['msg']:

            msg_type = self._convert_msg_kind(msg['type'])
            parsed_msg = self._get_msg_data(msg_type, msg['value'])

            if msg_type == 'reward':
                reward = rewards.pop(0) if rewards else None
                if reward:
                    parsed_msg['amount'] = int(reward[0]) * self.coef

            log = logs.pop(0)
            parsed_msg['is_error'] = not log['success']
            parsed_msg['kind'] = msg_type

            parsed_msgs.append(parsed_msg)

        return parsed_msgs

    @staticmethod
    def _convert_msg_kind(raw_type):
        return {
            'cosmos-sdk/MsgSend': 'transaction',
            'cosmos-sdk/MsgMultiSend': 'multisend',
            'cosmos-sdk/MsgWithdrawDelegationReward': 'reward',
            'cosmos-sdk/MsgDelegate': 'delegate',
            'cosmos-sdk/MsgBeginRedelegate': 'redelegate',
            'cosmos-sdk/MsgVote': 'vote'
        }.get(raw_type, raw_type)

    def _get_msg_data(self, msg_type, msg_value):
        # use original fields from api, change only amounts if needed
        msg_info = deepcopy(msg_value)

        if msg_type == 'reward':
            # amount for rewards are saved in different field,
            # see caller method
            pass

        elif msg_type == 'multisend':
            msg_result = self._process_msg_type_multisend(msg_value)
            if msg_result is not None:
                return msg_result
        else:
            amount_obj = msg_value['amount']
            amount_data = {}
            if isinstance(amount_obj, list):
                amount_data = amount_obj[0] if amount_obj else {}
            elif isinstance(amount_obj, dict):
                amount_data = amount_obj

            msg_info['amount'] = (int(amount_data['amount']) * self.coef
                                  if amount_data else None)

        return msg_info

    def _process_msg_type_multisend(self, msg_value):
        my_input = next((i for i in msg_value['inputs']
                         if i['address'].lower() == self.address.lower()),
                        None)

        my_output = next((i for i in msg_value['outputs']
                          if i['address'].lower() == self.address.lower()),
                         None)

        if my_input:
            to_address = (msg_value['outputs'][0]['address']
                          if len(msg_value['outputs']) == 1
                          else 'multiple')
            return {
                'from_address': self.address,
                'to_address': to_address,
                'amount': int(my_input['coins'][0]['amount']) * self.coef
            }
        if my_output:
            from_address = (msg_value['inputs'][0]['address']
                            if len(msg_value['inputs']) == 1
                            else 'multiple')
            return {
                'from_address': from_address,
                'to_address': self.address,
                'amount': int(my_output['coins'][0]['amount']) * self.coef
            }



import dateutil.parser

from blockapi.services import BlockchainAPI


class EosparkAPI(BlockchainAPI):
    """
    EOS
    API docs: https://developer.eospark.com/api-doc/https/
    Explorer: https://eospark.com
    """

    symbol = 'EOS'
    base_url = 'https://api.eospark.com/api'
    rate_limit = 0
    coef = 1e-6
    max_items_per_page = 20
    page_offset_step = 1

    supported_requests = {
        'get_balance': '?module=account&action=get_account_balance'
                       '&apikey={api_key}&account={address}',
        'get_token_balances': '?module=account&action=get_token_list'
                              '&apikey={api_key}&account={address}',
        'get_txs': '?module=account&action=get_account_related_trx_info'
                   '&apikey={api_key}&account={address}'
                   '&page={page}&size={size}'
    }

    def get_balance(self):
        """Not implemented yet"""
        pass

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        """Not implemented yet"""
        pass

    def parse_tx(self, tx):
        return {
            'symbol': tx['code'],
            'date': dateutil.parser.parse(tx['timestamp']),
            'from_address': tx['sender'],
            'to_address': tx['receiver'],
            'amount': tx['quantity'],
            # 'fee': fee,
            # 'gas_used': int(tx['gas_used']),
            # 'gas_limit': int(tx['gas_wanted']),
            'hash': tx['trx_id'],
            # 'block_hash': None,
            # 'block_num': tx['block_num'],
            'confirmed': True,
            'is_error': False,
            'type': 'normal' if tx['symbol'] == 'EOS' else 'token',
            'kind': 'transaction',
            'direction': 'outgoing' if self.address == tx['sender']
            else 'incoming',
            'raw': tx
        }



from pytest import mark

from blockapi.api.blockchair import (BlockchairGroestlcoinAPI,
                                     BlockchairBitcoinSvAPI)
from blockapi.test_init import test_addresses


class TestBlockchairAPI:
    GROESTL_ADDRESS = test_addresses['GRS'][0]
    BSV_ADDRESS = test_addresses['BSV'][0]

    @mark.vcr()
    def test_get_balance_groestl_coin(self):
        api = BlockchairGroestlcoinAPI(address=self.GROESTL_ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'GRS', 'amount': 330.00184162}]

    @mark.vcr()
    def test_get_balance_bitcoin_sv(self):
        api = BlockchairBitcoinSvAPI(address=self.BSV_ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'BSV', 'amount': 0.39608498000000003}]



from datetime import datetime

import pytz

from blockapi.services import BlockchainAPI, APIError


class NeoscanAPI(BlockchainAPI):
    """
    coins: neocoin
    API docs: https://neoscan.io/docs/index.html#api-v1
    Explorer: 
    """

    active = True

    symbol = 'NEO'
    base_url = 'https://api.neoscan.io/api/main_net/v1'
    rate_limit = 0
    coef = 1
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/get_balance/{address}',
        'get_txs': '/get_address_abstracts/{address}/{page}'
    }

    def __init__(self, address, api_key=None):
        super().__init__(address, api_key)
        paging_params = self.get_tx_paging_params()
        if paging_params is None:
            raise APIError('Can\'t initialize API paging.')
        else:
            self.max_items_per_page = self.page_offset_step = paging_params[1]
            self.total_txs_count = paging_params[2]

    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        for bal in response['balance']:
            if bal.get('asset_symbol') == self.symbol:
                return [{'symbol': self.symbol,
                         'amount': bal.get('amount') * self.coef}]

        return None

    def get_tx_paging_params(self):
        # total pages can be found on the first page
        response = self.request('get_txs',
                                address=self.address,
                                page=1)
        if 'total_pages' in response and 'page_size' in response:
            return (int(response['total_pages']),
                    int(response['page_size']),
                    int(response['total_entries']))
        else:
            return None

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        page = (offset // self.max_items_per_page) + 1
        # page_offset 'should be set' offset % self.max_items_per_page

        if limit > self.total_txs_count - offset:
            limit = self.total_txs_count - offset

        result = []

        while True:
            response = self.request('get_txs',
                                    address=self.address,
                                    page=page)

            if 'entries' in response:
                for t in response['entries']:
                    result.append(self.parse_tx(t))
                    limit -= 1
                    if limit == 0:
                        return result
            else:
                return result

            page += 1

    def parse_tx(self, tx):
        if tx['address_from'] == self.address:
            direction = 'outgoing'
        else:
            direction = 'incoming'

        return {
            'date': datetime.fromtimestamp(tx['time'], pytz.utc),
            'from_address': tx['address_from'],
            'to_address': tx['address_to'],
            'amount': tx['amount'],
            'fee': None,
            'hash': tx['txid'],
            'confirmed': None,
            'is_error': False,
            'type': 'normal',
            'kind': 'transaction',
            'direction': direction,
            'status': 'confirmed',
            'raw': tx
        }



import setuptools
from setuptools import find_packages

with open("README.md", "r") as f:
    long_description = f.read()

PACKAGES = find_packages(where='.')

setuptools.setup(
    name='blockapi',
    version='0.0.71',
    author='Devmons s.r.o.',
    description='BlockAPI library',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=PACKAGES,
    install_requires=[
        'requests>=2.22.0',
        'pytz>=2019.2',
        'python-dateutil>=2.8.0',
        'coinaddrng==1.0.25',
        'cfscrape>=2.0.8',
        'gevent>=1.4.0',
        'ethereum_input_decoder>=0.2.2',
        'web3>=5.2.2',
        'bs4>=0.0.1',
        'lxml>=4.4.1',
        'pytest',
        'pytest-vcr'
    ],
)




from datetime import datetime

import pytz

from blockapi.services import (
    AddressNotExist,
    BlockchainAPI,
    set_default_args_values
)


class DcrdataAPI(BlockchainAPI):
    """
    Decred
    API docs: https://github.com/decred/dcrdata#dcrdata-api
    Explorer: https://explorer.dcrdata.org
    """

    symbol = 'DCR'
    base_url = 'https://explorer.dcrdata.org/api'
    testnet_url = 'https://testnet.dcrdata.org/api'
    rate_limit = 0
    coef = 1
    max_items_per_page = 50000
    page_offset_step = max_items_per_page

    supported_requests = {
        'get_balance': '/address/{address}/totals',
        'get_transaction': '/tx/{tx_hash}?spends=true',
        'get_txs': '/address/{address}/count/{count}/skip/{skip}/raw'
    }

    def process_error_response(self, response):
        if response.status_code == 422:
            raise AddressNotExist()
        # else
        super().process_error_response(response)

    def get_balance(self):
        balance = self.request('get_balance', address=self.address)
        return [{'symbol': self.symbol, 'amount': balance['dcr_unspent']}]

    @set_default_args_values
    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        txs = self.request(
            'get_txs',
            address=self.address,
            count=limit,
            skip=offset
        )
        return [self.parse_tx(tx) for tx in txs]

    def get_tx(self, tx_hash):
        """Despite the name this method is not returning single tx
        but might result in the array of txs
        """
        tx = self.request('get_transaction', tx_hash=tx_hash)
        return self.parse_tx(tx)['result']

    def parse_tx(self, tx):
        kind = self.get_tx_kind(tx)

        parsed = {
            'transaction': self.parse_regular_tx,
            'ticket': self.parse_ticket,
            'vote': self.parse_vote,
            'revocation': self.parse_revocation
        }.get(kind)(tx)
        return {'kind': kind, 'result': parsed}

    @staticmethod
    def get_tx_kind(tx):
        tx_types = [t['scriptPubKey']['type'] for t in tx['vout']]
        if 'stakesubmission' in tx_types:  # and 'sstxcommitment', 'sstxchange'
            return 'ticket'
        elif 'stakegen' in tx_types:
            return 'vote'
        elif 'stakerevoke' in tx_types:
            return 'revocation'
        elif 'pubkeyhash' in tx_types:
            return 'transaction'  # 'regular' in api
        return None

    def parse_regular_tx(self, tx):
        # Tx in decred could contain several addresses, filter only mine
        ins = [v for v in tx['vin']
               if self.address in v.get('prevOut', {}).get('addresses', [])]
        outs = [o for o in tx['vout'] if self.address
                in o.get('scriptPubKey', {}).get('addresses', [])]

        # get_txs has time attribute, get_tx has block.time attribute
        if 'time' in tx:
            date = datetime.fromtimestamp(tx['time'], pytz.utc)
        else:
            date = datetime.fromtimestamp(tx['block']['time'], pytz.utc)

        parsed = []
        for i in ins:
            parsed.append({
                'date': date,
                'from_address': self.address,
                'to_address': None,  # multiple, TODO check it
                'amount': i['amountin'],
                'fee': None,
                'gas_limit': None,
                'hash': tx['txid'],
                'confirmed': tx['confirmations'] > self.confirmed_num,
                'is_error': False,
                'type': 'normal',
                'kind': 'transaction',
                'direction': 'outgoing',
                'status': None,
                'raw': tx
            })

        for o in outs:
            parsed.append({
                'date': date,
                'from_address': None,  # multiple, TODO check it
                'to_address': self.address,
                'amount': o['value'],
                'fee': None,
                'gas_limit': None,
                'hash': tx['txid'],
                'confirmed': tx['confirmations'] > self.confirmed_num,
                'is_error': False,
                'type': 'normal',
                'kind': 'transaction',
                'direction': 'incoming',
                'status': None,
                'raw': tx
            })
        return parsed

    @staticmethod
    def parse_ticket(tx):
        investment = sum(v['amountin'] for v in tx['vin'])
        ticket_cost = next((v['value'] for v in tx['vout']
                            if v['scriptPubKey']['type'] == 'stakesubmission'),
                           0)

        # pool fee is lower value then ticket cost,
        # but not sure if it's correct
        pool_fee = (min([v['amountin'] for v in tx['vin']])
                    if len(tx['vin']) > 1
                    else 0)

        return {
            'hash': tx['txid'],
            'purchased_on': datetime.fromtimestamp(tx['time'], pytz.utc),
            # all input tx ids in tx['vin'] should be same?
            'input_tx_id': tx['vin'][0]['txid'],
            'investment': investment,
            'ticket_cost': ticket_cost,
            'transaction_fee': investment - ticket_cost,
            'pool_fee': pool_fee,
            'block_hash': tx['blockhash'],
            'block_height': None,  # TODO
            'status': DcrdataAPI.get_ticket_status(tx)
        }

    @staticmethod
    def get_ticket_status(tx):
        # params for mainnet:
        # https://github.com/decred/dcrd/blob
        # /9da132b9823b20870122dc0bf795884cee99d922
        # /chaincfg/mainnetparams.go#L321
        if tx['confirmations'] < 1:
            return 'unmined'
        elif tx['confirmations'] < 256:
            return 'immature'
        # TODO - not sure about this one
        elif tx['confirmations'] < 7640:
            return 'live'
        elif tx['confirmations'] < 40960:
            return 'voted'
        elif tx['confirmations'] > 40960:
            return 'expired'

        # other statuses depends on specific txs
        # (votes -> voted, revocations -> revocated)
        # -> specific tx refers to ticket by ticket_hash
        # otherwise is live
        return 'live'

    @staticmethod
    def parse_vote(tx):
        total_input = 0
        reward = 0
        ticket_hash = None
        for v in tx['vin']:
            if 'stakebase' in v:
                reward = v['amountin']
                total_input += reward
            else:
                # value of ticket
                ticket_hash = v['txid']
                total_input += v['prevOut']['value']

        total_output = sum(v['value'] for v in tx['vout'] if v['value'])

        return {
            'hash': tx['txid'],
            'voted_on': datetime.fromtimestamp(tx['time'], pytz.utc),
            'spent_ticket_hash': ticket_hash,
            'reward': reward,
            'fee': total_input - total_output,
            'block_hash': tx['blockhash']
        }

    @staticmethod
    def parse_revocation(tx):
        ticket_hash = next((v['txid'] for v in tx['vin']), None)
        total_input = sum(v['amountin'] for v in tx['vin'] if v['amountind'])
        total_output = sum(v['value'] for v in tx['vout'] if v['value'])

        return {
            'hash': tx['txid'],
            'revocated_on': datetime.fromtimestamp(tx['time'], pytz.utc),
            'spent_ticket_hash': ticket_hash,
            'fee': total_input - total_output,
            'block_hash': tx['blockhash']
        }



from datetime import datetime

import pytz

from blockapi.services import (
    BlockchainAPI,
    APIError,
    AddressNotExist,
)


class CardanoExplorerAPI(BlockchainAPI):
    """
    Cardano
    API docs: https://cardanodocs.com/technical/explorer/api/
    Explorer: https://cardanoexplorer.com
    """

    symbol = 'ADA'
    base_url = 'https://cardanoexplorer.com/api'
    rate_limit = 0
    coef = 1e-6

    supported_requests = {
        'get_summary': '/addresses/summary/{address}'
    }

    def get_balance(self):
        summary = self._get_summary()
        retval = int(summary['Right']['caBalance']['getCoin']) * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        summary = self._get_summary()
        txs = summary['Right']['caTxList']
        return [self.parse_tx(t) for t in txs]

    def parse_tx(self, tx):
        my_input = next((i for i in tx['ctbInputs']
                         if i[0].lower() == self.address.lower()), None)

        my_output = next((i for i in tx['ctbOutputs']
                          if i[0].lower() == self.address.lower()), None)

        fee = None

        if my_input:
            fee = (int(tx['ctbInputSum']['getCoin'])
                   - int(tx['ctbOutputSum']['getCoin']))

            to_address = (tx['ctbInputs'][0][0]
                          if len(tx['ctbInputs']) else None)
            from_address = self.address
            amount = int(my_input[1]['getCoin']) * self.coef

        else:
            from_address = (tx['ctbOutputs'][0][0]
                            if len(tx['ctbOutputs']) else None)
            to_address = self.address
            amount = int(my_output[1]['getCoin']) * self.coef

        return {
            'date': datetime.fromtimestamp(tx['ctbTimeIssued'], pytz.utc),
            'from_address': from_address,
            'to_address': to_address,
            'amount': amount,
            'fee': fee,
            'gas': {},
            'hash': tx['ctbId'],
            'confirmed': None,
            'is_error': False,
            'type': 'normal',
            'kind': 'transaction',
            'direction': 'outgoing' if my_input else 'incoming',
            'raw': tx
        }

    def _get_summary(self):
        summary = self.request('get_summary', address=self.address)
        if summary.get('Left'):
            self._process_error(summary['Left'])
            return None

        return summary

    @staticmethod
    def _process_error(msg):
        if msg == 'Invalid Cardano address!':
            raise AddressNotExist()
        else:
            raise APIError(msg)



from pytest import mark

from blockapi.api.tzscan import TzscanAPI
from blockapi.test_init import test_addresses


class TestTzscanAPI:

    ADDRESS = test_addresses["XTZ"][0]
    REWARD_ADDRESS = test_addresses["XTZ"][1]

    def test_init(self):
        api = TzscanAPI(address=self.ADDRESS)
        assert api

    @mark.vcr()
    def test_get_balance(self):
        api = TzscanAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'XTZ', 'amount': 2068856.4582429999}]

    @mark.vcr()
    def test_get_rewards(self):
        api = TzscanAPI(address=self.ADDRESS)
        result = api.get_rewards()

        assert isinstance(result, list)
        # TODO: find better address, this one returns empty list

    @mark.vcr()
    def test_get_txs(self):
        api = TzscanAPI(address=self.ADDRESS)
        result = api.get_txs()

        assert isinstance(result, list)
        assert len(result) == 3

    @mark.vcr()
    def test_get_bakings_tzstats(self):
        api = TzscanAPI(address=self.REWARD_ADDRESS)
        result = api.get_bakings_tzstats()

        assert result["address"] == self.REWARD_ADDRESS
        assert "manager" in result
        assert "delegate" in result
        assert "manager_account" in result
        assert "delegate_account" in result

    @mark.vcr()
    def test_get_endorsements(self):
        api = TzscanAPI(address=self.ADDRESS)
        result = api.get_endorsements()

        assert isinstance(result, list)
        # TODO: find better address, this one returns empty list



from pytest import mark, raises

from blockapi.api.cosmos import CosmosAPI
from blockapi.services import InternalServerError, AddressNotExist

from blockapi.test_init import test_addresses


class TestCosmosAPI:

    ADDRESS = test_addresses["ATOM"][0]

    def test_init(self):
        api = CosmosAPI(address=self.ADDRESS)
        assert api

    def test_process_error_response(self):
        response = TestResponse()

        response.text = "Error"
        response.status_code = 500

        with raises(InternalServerError):
            CosmosAPI(address=self.ADDRESS).process_error_response(response)

        response.text = "Error decoding bech32 failed"

        with raises(AddressNotExist):
            CosmosAPI(address=self.ADDRESS).process_error_response(response)

    @mark.vcr()
    def test_get_info(self):
        api = CosmosAPI(address=self.ADDRESS)
        result = api.get_info()

        assert "result" in result
        assert "value" in result["result"]
        assert "address" in result["result"]["value"]
        assert "coins" in result["result"]["value"]
        assert isinstance(result["result"]["value"]["coins"], list)
        assert "denom" in result["result"]["value"]["coins"][0]
        assert "amount" in result["result"]["value"]["coins"][0]
        assert "account_number" in result["result"]["value"]
        assert "sequence" in result["result"]["value"]

    @mark.vcr()
    def test_get_balance(self):
        api = CosmosAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [{"symbol": "ATOM", "amount": 0.005959}]

    @mark.vcr()
    def test_get_incoming_txs(self):
        api = CosmosAPI(address=self.ADDRESS)
        api.get_incoming_txs()

        # TODO: provider is not sending correct data

    @mark.vcr()
    def test_get_outgoing_txs(self):
        api = CosmosAPI(address=self.ADDRESS)
        api.get_outgoing_txs()

        # TODO: provider is not sending correct data


class TestResponse:

    text = None
    status_code = 200



import re
from time import sleep

import cfscrape
import requests
from bs4 import BeautifulSoup
from ethereum_input_decoder import AbiMethod
from web3 import Web3

from blockapi.api import EtherscanAPI


class Ethereum:
    def __init__(self, node_url, etherscan_api_key):
        self.node_url = node_url
        self.etherscan_api_key = etherscan_api_key
        self.web3 = Web3(Web3.HTTPProvider(self.node_url))
        self.abi = None

    def load_abi(self, contract):
        myapi = EtherscanAPI(contract, api_key=self.etherscan_api_key)
        self.abi = myapi.get_abi(contract)['result']

    def to_checksum_addr(self, address):
        return self.web3.toChecksumAddress(address)

    def get_contract(self, contract):
        # fallback to automatic loading if ABI is not set from outside
        if self.abi is None:
            self.load_abi(contract)
        return self.web3.eth.contract(address=Web3.toChecksumAddress(
            contract), abi=self.abi)

    def get_tx_by_hash(self, txhash):
        tx = self.web3.eth.getTransaction(txhash)
        return tx

    def get_function_by_inputdata(self, tx_input):
        tx_input_decoded = AbiMethod.from_input_lookup(
            bytes.fromhex(tx_input[2:]))
        tx_input_values = list(tx_input_decoded.values())
        tx_function = tx_input_values[0]
        return tx_function

    def get_erc20_balances(self, address, tokens: dict):
        addr = '0xD1f89108Ba9525a11BEAf8bdC5058B4946a6C501'
        addr = Web3.toChecksumAddress(addr)
        address = Web3.toChecksumAddress(address)

        tokens_ch = []
        for token in tokens:
            tokens_ch.append(
                Web3.toChecksumAddress(tokens[token]['contract_address']))

        bal_sc = self.get_contract(addr)
        balances = bal_sc.functions.batchTokenBalances([address],
                                                       tokens_ch).call()

        decimals = [tokens[token]['decimals'] for token in tokens]
        symbols = [token for token in tokens]
        contract_addresses = [tokens[token]['contract_address']
                              for token in tokens]

        return [{'amount': float(b)*pow(10,-d),
                 'symbol': symbol, 'contract_address': sc }
                for b,d, symbol, sc in
                zip(balances,decimals, symbols, contract_addresses) if b > 0]


class Infura(Ethereum):
    def __init__(self, network, api_key, etherscan_api_key):
        self.network = network
        self.api_prefix = network if network != "mainnet" else "api"
        self.api_key = api_key
        self.etherscan_api_key = etherscan_api_key
        self.infura_url = 'https://{}.infura.io/v3/{}'.format(self.network,
                                                              self.api_key)
        super().__init__(self.infura_url, etherscan_api_key)


class ERC20Token:
    def __init__(self):
        self.url = 'https://etherscan.io/tokens?p={}'
        self.token_url = 'https://etherscan.io/token/{}'
        self.page = 0
        self.reqobj = cfscrape.create_scraper()
        self.tokens = {}

    def get_token_list(self):
        """
        Scrapes the ERC20 token list from etherscan.io

        :return: dictionary containing currency_name, contract_addres,
                 price, change, volume, market_cap, holders - for each token
        :rtype: dict
        """
        scrape_result = True
        result_msg = ''
        page = 0

        while True:
            status_code, rows = self._get_table_rows(page)
            if status_code != 200:
                scrape_result = False
                result_msg = 'error {} on page {}'.format(status_code,
                                                          page)
                break

            if rows is None:
                break

            for row in rows:
                result = self._parse_table_row(row)
                decimals = self._get_token_details(result[2])

                self.tokens[result[0]] = \
                    {'currency_name': result[1],
                     'contract_address': result[2],
                     'price': float(result[3]),
                     'change': result[4],
                     'volume': result[5],
                     'market_cap': result[6],
                     'holders': result[7],
                     'decimals': decimals}

            page += 1

        return {'result': scrape_result, 'result_msg': result_msg,
                'tokens': self.tokens}

    def _get_token_details(self, contract_address):
        token_details_url = self.token_url.format(contract_address)
        i = 1
        while True:
            try:
                result = self.reqobj.get(token_details_url)
                if result.text.find('Sorry, You have reached') != -1:
                    i += 1
                    sleep(0.3 * i)
                    continue
                break
            except requests.exceptions.ChunkedEncodingError:
                i += 1

            sleep(0.3 * i)

        if result.status_code != 200:
            return None

        soup = BeautifulSoup(result.text, "lxml")
        decimals_div = soup.body.find("div",
                                      {"id": "ContentPlaceHolder1_trDecimals"})
        if decimals_div is None:
            return None

        decimals_class = decimals_div.find("div", {"class": "col-md-8"})
        return int(decimals_class.text)

    def _get_table_rows(self, page):
        erc20_url = self.url.format(page)
        i = 1
        while True:
            try:
                result = self.reqobj.get(erc20_url)
                if result.text.find('Sorry, You have reached') != -1:
                    i += 1
                    sleep(0.3 * i)
                    continue
                break
            except requests.exceptions.ChunkedEncodingError:
                i += 1

            sleep(0.3 * i)

        if result.status_code == 200:
            soup = BeautifulSoup(result.text, "lxml")
            table = soup.body.find("table", {"id": "tblResult"})

            if table is None:
                return result.status_code, None

            table_body = table.find_all('tbody')[0]
            table_rows = table_body.find_all('tr')

            return result.status_code, table_rows
        else:
            return result.status_code, None

    def _parse_table_row(self, row):
        row_ahref = row.find_all('a')[0]
        cointext = row_ahref.text

        coin_sc = row_ahref['href'][7:]
        currency_symbol = self._get_currency_symbol(cointext)
        currency_name = self._get_currency_name(cointext)

        row_tds = row.find_all('td')
        row_tds[2].find('div').decompose()

        currency_price = ERC20Token._get_number(row_tds[2].text,
                                                float)
        currency_change = ERC20Token._get_number(row_tds[3].text,
                                                 float)
        currency_volume = ERC20Token._get_number(row_tds[4].text, int)
        market_cap = ERC20Token._get_number(row_tds[5].text, int)
        holders = ERC20Token._get_number(row_tds[6].text, int)

        return (currency_symbol, currency_name, coin_sc, currency_price,
                currency_change, currency_volume, market_cap, holders)

    def _get_currency_symbol(self, cointext):
        try:
            currency_symbol = re.search('\([A-Za-z0-9]+\)',
                                        cointext).group(0)[1:-1]
        except AttributeError:
            currency_symbol = cointext

        return currency_symbol

    def _get_currency_name(self, cointext):
        currency_name = re.sub('\([A-Za-z0-9]+\)', '',
                               cointext).strip()
        return currency_name

    @staticmethod
    def _get_number(num_string, rtype):
        """
        converts string in currency format $1234,56.78  or 1234.56 % into
        a user-defined type (usually float or int)

        :param num_string: string in US currency format / percentage
        :type num_string: string
        :param rtype: class to convert the resulting number to
        :type rtype: class
        :return: number converted from the string to the desired type
        """
        try:
            result = rtype(
                re.sub('[$,%]', '', num_string))
        except ValueError:
            result = None

        return result

    def get_contract_by_symbol(self, symbol):
        if len(self.tokens) == 0:
            return None

        if self.tokens.get(symbol) is not None:
            return self.tokens[symbol]['contract_address']
        else:
            return None
